/*
   piktc_svc.c -- PIKT master control program (piktc) service daemon

   $Header: /home/cvs/ucgsb/pikt/src/piktc/piktc_svc.c,v 1.30 2000/02/13 05:34:58 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA

   --------------------------------------------------------------------------

   Usage:  piktc_svc [-GhT]

           -G              run in debug mode
           -h              show program help
           -T              run in test mode

 */

#include "../lib/common.h"
#include "piktc.h"

const char piktc_version[] = "PIKT piktc_svc version " VERSION;

/* command-line options */
/* int ischeck = FALSE; */
int isdebug = FALSE;
/* int isdelete = FALSE; */
/* int isdoublecheck = FALSE; */
/* int isdiff = FALSE; */
/* int isdisable = FALSE; */
/* int isenable = FALSE; */
int isexec = FALSE;
int isexecwait = FALSE;
int ishelp = FALSE;
/* int isinstall = FALSE; */
/* int iskill_piktd = FALSE; */
/* int iskill_piktc_svc = FALSE; */
/* int islist = FALSE; */
/* int isrestart_piktd = FALSE; */
/* int isrestart_piktc_svc = FALSE; */
int istest = FALSE;
/* int isshow = FALSE; */
/* int isverbose = FALSE; */

int bailout = FALSE;	/* signifies command-line error */

/* PIKT.conf options */
int uid = 0;		/* default to root */
char *master = (char *) NULL;
char *domain = (char *) NULL;
char *accessmode = (char *) NULL;
char *master_address = (char *) NULL;
int isauth_by_master = TRUE;
int isauth_by_master_address = FALSE;
int rpc_timeout = RPC_TIMEOUT;
int rpc_retry_timeout = RPC_RETRY_TIMEOUT;
int rpcinfo_retries = RPCINFO_RETRIES;
int isverboselog = FALSE;
int isautosyslog = FALSE;
int istcponly = FALSE;
int execcmdout_limit = EXECCMDOUT_LIMIT;

int lineno = 0;

char hostname[MAXIDENT+1];
char domainname[MAXIDENT+1];

int ismaster = FALSE;

char *hostserv;
char *rqstserv;
int statserv;

unsigned long services = ALL_SERVICES;	/* all services enabled by default */

int
main (int argc, char *argv[])
{
	static char _main_[] = "main";
	pid_t pid;
	static char pidstr[8];

	init(argc, argv, (void (*)(void))piktc_svc_usage);
	if (bailout)
		exit(1);

	if (! isfile(FILECMD, TRUE) || ! isfile(PSCMD, TRUE)) {
		/* error(s) logged in PIKTLOG */
		exit(1);
	}

	/* apply the lock, to prevent other concurrent piktc_svc runs */
	if (islock(piktc_svclock)) {
		fprintf(stderr, "%s found!  Another piktc session may be in progress.\nTry again later (else rm the file).\n", piktc_svclock);
		fflush(stderr);
		exit(1);
	} else {
		pid = getpid();
		/* we add one to the pid, because below when we call
		   piktc_svc_main(), the current process forks a child,
		   which will *probably* (although not necessarily) have
		   the next highest pid
		 */
		sprintf(pidstr, "%u", (unsigned)pid+1);
		if (mklock(piktc_svclock, pidstr, TRUE, TRUE) == ERROR)
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in main(), mklock() error for %s", piktc_svclock);
	}

	logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "daemon (re)start");

	/* more initializations:  no initial service registrations */
	hostserv = mystrdup("", _main_);
	rqstserv = mystrdup("", _main_);

	piktc_svc_main();	/* go into daemon mode */

	exit(0);	/* never reached */
}

/* output program usage to screen
 */
void
piktc_svc_usage (void)
{
	printf("\nUsage:  piktc_svc [-GhT]\n\n");
	printf("\t-G\t\t\trun in debug mode\n");
	printf("\t-h\t\t\tshow program help\n");
	printf("\t-T\t\t\trun in test mode\n\n");
	return;
}
