/*
   rpcsvc.c -- PIKT rpc service (piktc_svc) routines

   $Header: /home/cvs/ucgsb/pikt/src/piktc/rpcsvc.c,v 1.54 2000/02/21 00:15:53 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998-2000 by Robert Osterlund.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   This product includes software developed by the University of
   California, Berkeley and its contributors.

   Send problem reports, bouquets and brickbats to:

   robert.osterlund@gsb.uchicago.edu

   Robert Osterlund
   Graduate School of Business
   University of Chicago
   1101 E. 58th Street, W309
   Chicago, IL 60637
   USA
 */

#define PORTMAP

#include "../lib/common.h"

#include <rpc/rpc.h>

#if defined(HAVE_NETCONFIG_H)
#include <netconfig.h>
#endif

#include <rpc/xdr.h>
#include <rpc/clnt.h>
#include "rpc.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <sys/utsname.h>
#include <dirent.h>

#if defined(HAVE_SYS_SYSTEMINFO_H)
	#include <sys/systeminfo.h>
#endif

#include "piktc.h"

#ifdef NO_RPC_SVC
#define LIST_FILE_1_SVC			list_file_1
#define INSTALL_FILE_1_SVC		install_file_1
#define FETCH_FILE_1_SVC		fetch_file_1
#define DELETE_FILE_1_SVC		delete_file_1
#define EDIT_PIKTD_CONF_1_SVC		edit_piktd_conf_1
#define EXECUTE_COMMAND_1_SVC		execute_command_1
#define EXECUTE_COMMANDWAIT_1_SVC	execute_commandwait_1
#define FREE_EXECCMDOUT_1_SVC		free_execcmdout_1
#define EXECUTE_ALERT_1_SVC		execute_alert_1
#define SHOW_ALERT_STATUS_1_SVC		show_alert_status_1
#define KILL_PIKTD_1_SVC		kill_piktd_1
#define KILL_PIKTC_SVC_1_SVC		kill_piktc_svc_1
#define RESTART_PIKTD_1_SVC		restart_piktd_1
#define RESTART_PIKTC_SVC_1_SVC		restart_piktc_svc_1
#define REGISTER_REQUEST_1_SVC		register_request_1
#define CALL_BACK_1_SVC			call_back_1
#else
#define LIST_FILE_1_SVC			list_file_1_svc
#define INSTALL_FILE_1_SVC		install_file_1_svc
#define FETCH_FILE_1_SVC		fetch_file_1_svc
#define DELETE_FILE_1_SVC		delete_file_1_svc
#define EDIT_PIKTD_CONF_1_SVC		edit_piktd_conf_1_svc
#define EXECUTE_COMMAND_1_SVC		execute_command_1_svc
#define EXECUTE_COMMANDWAIT_1_SVC	execute_commandwait_1_svc
#define FREE_EXECCMDOUT_1_SVC		free_execcmdout_1_svc
#define EXECUTE_ALERT_1_SVC		execute_alert_1_svc
#define SHOW_ALERT_STATUS_1_SVC		show_alert_status_1_svc
#define KILL_PIKTD_1_SVC		kill_piktd_1_svc
#define KILL_PIKTC_SVC_1_SVC		kill_piktc_svc_1_svc
#define RESTART_PIKTD_1_SVC		restart_piktd_1_svc
#define RESTART_PIKTC_SVC_1_SVC		restart_piktc_svc_1_svc
#define REGISTER_REQUEST_1_SVC		register_request_1_svc
#define CALL_BACK_1_SVC			call_back_1_svc
#endif

CLIENT *clnt;

req reqserv;

static execres xres;

int waitpid_options = 0
#ifdef NEED_WAITPID_WNOHANG
                        | WNOHANG
#endif
                        ;

int chkrequest (ritm *, rfil *, struct svc_req *, const char *);
void free_execcmdout_svc(void);
pid_t daemonpid(void);

/* for all below, return 0 on success, else ERROR, NONE, DENIED, ...;
   log all errors including errno info, if available, in the local piktc.log
 */

/* first, check the PIKT.conf service access parameters to see if this request
   is authorized; if not, deny the request

   second, attempt to authenticate the requesting piktc client system;
   compare the client name and/or client address to the approved master
   and/or master address; if they match, grant authentication; in special
   cases, make further attempts to grant authentication

   third, if access is set to local, check to see that the client's network
   domain matches the local network domain; if they dont' match, deny
   authentication

   fourth, check to see that the requesting account is an authorized account;
   if not, deny authentication

   fifth, and if failed authentication hasn't already denied the request,
   do a call_back, to see if this request is registered on the master;
   if not, deny the request

   return values:  0 (approved), UNAUTHORIZED, DENIED
 */
int
chkrequest (ritm *an, rfil *rf, struct svc_req *rqstp, const char *request)
{
	int res;
	struct sockaddr_in *sock_in;
	struct hostent *hp;
	char client_name[MAXIDENT+1];
	char *cp;
	char *client_domain;
	char *hap;
	unsigned long clntaddr;
	char *halp;
	char client_address[128];	/* dotted-decimal string; be generous
					   to accommodate IPv6/IPng
					 */
	struct authunix_parms *unix_cred;
	char *aup_client_name;
	int aup_client_uid;
	int authenticated;

	res = 0;

	/* always approve register_request */
	if (! strcmp(request, "register_request"))
		return res;

	authenticated = FALSE;

	aup_client_name = "<<ERR>>";
	aup_client_uid = -1;

	/* first, check the PIKT.conf service access parameters to see if
	   this request is authorized
	 */

	if (services != ALL_SERVICES) {  /* one or more services disabled */
	
		if (isdebug) {
			logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in chkrequest(), request is %s, services is %x", request, services);
		}

		/* call_back is always authorized
		if (! strcmp(request, "call_back")) {
			;

		} else
		 */

		if (! strcmp(request, "execute_command")) {
			if (! (services & EXECUTE_COMMANDS_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "execute_alert")) {
			if (! (services & EXECUTE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "install_file")) {
			if (! strcmp(rf->dir, altdir) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & INSTALL_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, objdir) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & INSTALL_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, prgdir) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & INSTALL_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & INSTALL_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "edit_piktd_conf")) {
			if (((an->state == ADD) || (an->aux && *(an->aux))) &&
			    ! (services & ENABLE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (((an->state == DELETE) || (an->aux && ! *(an->aux))) &&
			           ! (services & DISABLE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "show_alert_status")) {
			if (! (services & SHOW_ALERTS_STATUS_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "list_file")) {
			if (! strcmp(rf->dir, altdir) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & LIST_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, objdir) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & LIST_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, prgdir) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & LIST_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & LIST_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "fetch_file")) {
			if (! strcmp(rf->dir, altdir) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & FETCH_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, objdir) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & FETCH_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, prgdir) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & FETCH_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & FETCH_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "delete_file")) {
			if (! strcmp(rf->dir, altdir) &&
			    ! strcmp(rf->ext, ".alt") &&
			    ! (services & DELETE_ALERTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! strcmp(rf->dir, objdir) &&
			           ! strcmp(rf->ext, ".obj") &&
			           ! (services & DELETE_OBJECTS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if ((! strcmp(rf->dir, prgdir) ||
			            ! strcmp(rf->ext, "NIL")) &&
			           ! (services & DELETE_PROGRAMS_SERVICE)) {
				res = UNAUTHORIZED;
			} else if (! *(rf->ext) &&
			           ! (services & DELETE_FILES_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "kill_piktd")) {
			if (! (services & KILL_PIKTD_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "kill_piktc_svc")) {
			if (! (services & KILL_PIKTC_SVC_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "restart_piktd")) {
			if (! (services & RESTART_PIKTD_SERVICE)) {
				res = UNAUTHORIZED;
			}

		} else if (! strcmp(request, "restart_piktc_svc")) {
			if (! (services & RESTART_PIKTC_SVC_SERVICE)) {
				res = UNAUTHORIZED;
			}
		}

		if (res == UNAUTHORIZED) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in chkrequest(), %s request is unauthorized", request);
			return res;
		}
	}

	/* gather the client info */

	sock_in = svc_getcaller(rqstp->rq_xprt);

	/* get the client host's (i.e., the would-be control machine's)
	   fully qualified domain name
	 */
	hp = gethostbyaddr((char *) &sock_in->sin_addr, sizeof(sock_in->sin_addr), AF_INET);
	strcpy(client_name, hp->h_name);
	client_domain = (cp = strchr(client_name, '.')) ? cp+1 : cp;
	while (! client_domain && (hap = *(hp->h_aliases))) {
		strcpy(client_name, hap);
		hp->h_aliases++;
		client_domain = (cp = strchr(client_name, '.')) ? cp+1 : cp;
	}

	/* convert client address to host byte order */
	clntaddr = ntohl(sock_in->sin_addr.s_addr);

	/* get the dotted decimal form of the client address */
	strcpy(client_address, inet_ntoa(sock_in->sin_addr));

	/* get authunix credentials, including client_name & client_uid */
	unix_cred = (struct authunix_parms *) rqstp->rq_clntcred;
	aup_client_name = unix_cred->aup_machname;
	aup_client_uid = unix_cred->aup_uid;

	/* second, attempt to authenticate this client */

	if (strcmp(request, "call_back")) {

		/* typical cases */
		if (isauth_by_master && isauth_by_master_address) {
			if (! strcmp(client_name, master)           &&
			    ! strcmp(client_address, master_address)  ) {
				authenticated = TRUE;
			}
		} else if (isauth_by_master) {
			if (! strcmp(client_name, master)) {
				authenticated = TRUE;
			}
		} else if (isauth_by_master_address) {
			if (! strcmp(client_address, master_address)) {
				authenticated = TRUE;
			}
		}

		/* special cases */
		if (! authenticated) {
			/* if is validation self test */
			if (istest) {
				if (hostmatch(client_name, master)) {
					authenticated = TRUE;
				}
			/* else if is fetch_file */
			} else if (! strcmp(request, "install_file") &&
			           ! strcmp(rqstserv, "fetch_file")  &&
		        	   hostmatch(client_name, hostserv)   ) {
				authenticated = TRUE;
			/* else the client_name is not a fully qualified
			   host name, so cannot match master (which must be
			   a fully qualified host name); we will make one
			   last attempt to authenticate this client by
			   matching addresses
			 */
			 } else if (! client_domain            &&
			            ! isauth_by_master_address   ) {
				hp = gethostbyname(master);
				while (! authenticated              &&
				       (halp = *(hp->h_addr_list))   ) {
					if (memcmp(halp, &sock_in->sin_addr, sizeof(sock_in->sin_addr)) == 0) {
						authenticated = TRUE;
					}
					hp->h_addr_list++;
				}
			}
		}

		#ifdef REALSECURE
		if (authenticated) {
			/* this aup_client_name check is somewhat redundant,
			   given the earlier check of client_name
			 */
			if (! aup_client_name || ! strlen(aup_client_name) || strncasecmp(aup_client_name, master, strlen(aup_client_name))) {
				authenticated = FALSE;
			}
		}
		#endif

	} else {	/* is a call_back */

		/* for now, accept as valid a callback from any host
		   (but note that we still check the domain below, if
		   access is set to 'local', also the uid); in the future,
		   we might want to check against the systems list in
		   systems.cfg
		 */
		authenticated = TRUE;

	}

	/* third, check the domain access */

	if (! strcasecmp(accessmode, "local")) {
		/* restrict requests to within the local domain only,
		   possibly setting authenticated to FALSE;
		   note that this check is also done for call_back
		 */
		if (! strcasecmp(domain, "domain")) {
			if (! client_domain || (! strstr(client_domain, domainname) && ! strstr(domainname, client_domain))) {
				authenticated = FALSE;
			}
		} else {	/* there is an actual specified domain */
			if (! client_domain || strcasecmp(client_domain, domain)) {
				authenticated = FALSE;
			}
		}
	} /* else accessmode is "global", so don't check anything */

	/* fourth, check the client uid */

	/* note that the uid check is done also in the case of
	   call_back, so the uid in the master and slave PIKT.conf
	   have to match!
	 */
	if (! istest && (aup_client_uid != uid)) {
		authenticated = FALSE;
	}

	if (isdebug) {		/* show client authentication stuff */

		/* from PIKT.conf */
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "uid (PIKT.conf uid) is %d", uid);
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "master (PIKT.conf master) is %s", master ? master : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "master_address (PIKT.conf master_address) is %s", master_address ? master_address : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "domain (PIKT.conf domain) is %s", domain ? domain : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "access (PIKT.conf access) is %s", accessmode ? accessmode : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "authenticate_by_master (PIKT.conf authenticate_by_master) is %s", isauth_by_master ? "YES" : "NO");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "authenticate_by_master_address (PIKT.conf authenticate_by_master_address) is %s", isauth_by_master_address ? "YES" : "NO");

		/* from local host */
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "local hostname (hostname) is %s", hostname ? hostname : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "local domainname (domainname) is %s", domainname ? domainname : "unknown");

		/* from client */
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "client uid (aup_client_uid) is %d", aup_client_uid);
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "client name (client_name) is %s", client_name ? client_name : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "client name (aup_client_name) is %s", aup_client_name ? aup_client_name : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "client address (sock_in->sin_addr.s_addr) is %x", sock_in->sin_addr.s_addr);
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "client address (clntaddr) is %x", clntaddr);
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "client address (client_address) is %s", client_address ? client_address : "unknown");
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "client domain (client_domain) is %s", client_domain ? client_domain : "unknown");

	}

	if (authenticated) {
		if (strcmp(request, "call_back") || isverboselog)
			logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "%s request from uid %d on client %s (address %s)", request, aup_client_uid, client_name, client_address);
		res = 0;	/* should still be 0 */
	} else {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in chkrequest(), unauthenticated %s request from uid %d on client %s (address %s)", request, aup_client_uid, client_name, client_address);
		res = DENIED;
		return res;
	}

	/* fifth, do the call_back, to see if this request is registered on
	   the master; note: we don't call_back a call_back request!
	 */

	/* if call_back has not been turned off, and if this request is not a
	   call_back request, do a call_back
	 */
	if (IS_CALL_BACK && strcmp(request, "call_back")) {

		if (ismaster) {
			/* in effect, we are doing here what we would do if
			   doing a call_back() except looking entirely locally
			 */
			if (hostserv && rqstserv && hostmatch(hostname, hostserv)) {
				if (! strcmp(request,  rqstserv)) {
					res = 0;  /* should still be 0 */
				/* the enable_alerts special case */
				} else if (! strcmp(request, "list_file") &&
					   ! strcmp(rqstserv, "edit_piktd_conf")) {
					res = 0;  /* should still be 0 */
				} else {
					res = DENIED;
				}
			/* the fetch_file special case */
			} else if (! strcmp(request, "install_file") &&
			           ! strcmp(rqstserv, "fetch_file")) {
				res = 0;  /* should still be 0 */
			} else if ((statserv == EXECUTE) && (! strcmp(request, "execute_command") || ! strcmp(request, "execute_alert"))) {
				if (! strcmp(request, rqstserv) || ! *rqstserv) {
					res = 0;
				}
			} else {
				res = DENIED;
			}
			if (isdebug || (res == DENIED))
				logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in chkrequest(), hostserv is %s, rqstserv is %s, statserv is %d, hostname is %s, request is %s", hostserv, rqstserv, statserv, hostname, request);
		}
		/* no call_back for systems where NO_RPC_SVC is defined */
		#ifndef NO_RPC_SVC
		else {
			res = call_back(hostname, request, FALSE);
		}
		#endif
	}	/* else nothing:  we don't call_back a call_back! */

	/* finally, return the result */

	return res;
}

/* register a service request (the request and the hostname that the
   request is sent to);
   just to be safe, always service a register_request, whether IS_CALL_BACK
   or not
 */
int *
REGISTER_REQUEST_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static char _REGISTER_REQUEST_1_SVC_[] = "REGISTER_REQUEST_1_SVC";
	static int res;

	if ((res = chkrequest((ritm *)NULL, (rfil *)NULL, rqstp, "register_request")) != 0)
		return &res;

	FREE(hostserv);
	FREE(rqstserv);
	hostserv = mystrdup(rq->hostname, _REGISTER_REQUEST_1_SVC_);
	rqstserv = mystrdup(rq->request, _REGISTER_REQUEST_1_SVC_);
	statserv = rq->state;

	if (isdebug)
		logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in REGISTER_REQUEST_1_SVC(), hostserv is %s, rqstserv is %s, statserv is %d", hostserv, rqstserv, statserv);

	return &res;
}

#ifndef NO_RPC_SVC

/* compare the request and hostname of the calling back piktc_svc with
   the locally registered request and hostname; if they match, return okay;
   return values: 0 (approved), DENIED; (chkrequest() can return
   UNAUTHORIZED, but not in the case of call_back, which is always
   authorized
 */
int *
CALL_BACK_1_SVC (rqst *rq, struct svc_req *rqstp)
{
	static int res;

	if (IS_CALL_BACK) {

		if ((res = chkrequest((ritm *)NULL, (rfil *)NULL, rqstp, "call_back")) != 0)
			return &res;

		if (isdebug)
			logreturn(PIKTLOG, LOG_DEBUG, __FILE__, __LINE__, FALSE, "in CALL_BACK_1_SVC(), hostserv is %s, rqstserv is %s, statserv is %d, rq->hostname is %s, rq->request is %s", hostserv, rqstserv, statserv, rq->hostname, rq->request);

		/* when doing a 'piktc -X' (execute with no wait) to many
		   different clients, all of which are doing callbacks
		   asynchronously, the currently registered hostserv might
		   not match the hostname of the machine doing the callback,
		   so in this case only we relax the requirement that the
		   local hostserv and the rq->hostname match; a match is
		   still required for 'piktc -x' (execute with wait) and
		   all other services
		 */
		if (hostserv && rqstserv && hostmatch(rq->hostname, hostserv)) {
			if (! strcmp(rq->request, rqstserv)) {
				res = 0;
			/* the enable_alerts special case */
			} else if (! strcmp(rq->request, "list_file") &&
			           ! strcmp(rqstserv, "edit_piktd_conf")) {
				res = 0;
			} else {
				res = DENIED;
			}
		} else if ((statserv == EXECUTE) && (! strcmp(rq->request, "execute_command") || ! strcmp(rq->request, "execute_alert"))) {
			if (! strcmp(rq->request, rqstserv) || ! *rqstserv) {
				res = 0;
			}
		} else {
			res = DENIED;
		}

	} else {

		res = 0;	/* if call_back checks have been turned off,
				   always return success
				 */
	}

	return &res;
}

#endif	/* #ifndef NO_RPC_SVC */

/* report if a file exists
 */
int *
LIST_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char *fn;
	struct stat fbuf;
	static int res;

	if ((res = chkrequest((ritm *)NULL, rf, rqstp, "list_file")) != 0)
		return &res;

	fn = (char *) mymalloc((strlen(rf->dir)+strlen(rf->file)+strlen(rf->ext)+2) * sizeof(char), "LIST_FILE_1_SVC");
	sprintf(fn, "%s/%s%s", rf->dir, rf->file, strcmp(rf->ext, "NIL") ? rf->ext : "");
	res = stat(fn, &fbuf);	/* stat returns 0 on success,
				   -1 (ERROR) on failure
				 */
	free(fn);
	return &res;
}

/* install a file
 */
int *
INSTALL_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char *fn;
	FILE *fp;
	mode_t mode;
	uid_t uid;
	gid_t gid;

	static int res;

	if ((res = chkrequest((ritm *)NULL, rf, rqstp, "install_file")) != 0)
		return &res;

	fn = (char *) mymalloc((strlen(rf->dir)+strlen(rf->file)+strlen(rf->ext)+2) * sizeof(char), "INSTALL_FILE_1_SVC");
	sprintf(fn, "%s/%s%s", rf->dir, rf->file, strcmp(rf->ext, "NIL") ? rf->ext : "");
	if ((fp = fopen(fn, "w")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in install_file(), fopen() error for %s", fn);
		free(fn);
		res = ERROR;
		return &res;
	}
	res = fwrite(rf->text, sizeof(char), rf->size, fp);
	if (res < rf->size) {
		if (res == 0)
			res = errno;
		else
			res = ERROR;
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, res == ERROR ? 0 : 1, "in install_file(), fwrite() error for %s", fn);
		if (fclose(fp)) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in install_file(), fclose() error for %s", fn);
			free(fn);
			res = ERROR;
			return &res;
		}
		return &res;
	}
	res = fclose(fp);
	if (res != 0) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in install_file(), fclose() error for %s", fn);
		free(fn);
		res = ERROR;
		return &res;
	}
	mode = (mode_t) strtol(rf->mode, NULL, 8);	/* octal */
	if ((res = chmod(fn, mode)) != 0) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in install_file(), chmod() error for %s", fn);
		free(fn);
		res = errno;
		return &res;
	}
	uid = (uid_t) atoi(rf->uid);
	gid = (gid_t) atoi(rf->gid);
	if ((res = chown(fn, uid, gid)) != 0) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in install_file(), chown() error for %s", fn);
		free(fn);
		res = errno;
		return &res;
	}
	free(fn);
	return &res;
}

/* fetch a file
 */
int *
FETCH_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static int res;

	if ((res = chkrequest((ritm *)NULL, rf, rqstp, "fetch_file")) != 0)
		return &res;

	res = ! install_file(master, rf->file, rf->dir, rf->ext, MODE_FIL_DEFAULT, UID_DEFAULT, GID_DEFAULT, FALSE);
	return &res;
}

/* delete a file
 */
int *
DELETE_FILE_1_SVC (rfil *rf, struct svc_req *rqstp)
{
	static char *fn;
	struct stat fbuf;
	static int res;

	if ((res = chkrequest((ritm *)NULL, rf, rqstp, "delete_file")) != 0)
		return &res;

	fn = (char *) mymalloc((strlen(rf->dir)+strlen(rf->file)+strlen(rf->ext)+2) * sizeof(char), "DELETE_FILE_1_SVC");
	sprintf(fn, "%s/%s%s", rf->dir, rf->file, strcmp(rf->ext, "NIL") ? rf->ext : "");
	if (stat(fn, &fbuf) == 0) {
		res = remove(fn);	/* remove returns 0 on success */
		if (res == ERROR) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in delete_file(), remove() error for %s", fn);
			res = errno;
		}
	} else	/* file not found, so signify success in removing it */
		res = 0;
	free(fn);
	return &res;
}

/* for both enablement and disablement, for the given alert, remove its
   line from the piktd.conf file (if present); for enablement, add a new
   line to that file for the given alert
 */
int *
EDIT_PIKTD_CONF_1_SVC (ritm *an, struct svc_req *rqstp)
{
	FILE *pcfp, *tpcfp;
	char line[MAXLINE+1];
	char target[MAXLINE+1];
	static int res;

	if ((res = chkrequest(an, (rfil *)NULL, rqstp, "edit_piktd_conf")) != 0)
		return &res;

	if ((pcfp = fopen(piktdconf, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, TRUE, "in edit_piktd_conf(), fopen() error for %s", piktdconf);
		/* this might be a brand new installation, hence no prior
		   piktd.conf file
		 */
		/*
		res = ERROR;
		return &res;
		 */
	}
	if ((tpcfp = fopen(piktdconftmp, "w")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in edit_piktd_conf(), fopen() error for %s", piktdconftmp);
		res = ERROR;
		return &res;
	}
	if (pcfp) {
		sprintf(target, "+A %s", an->name);
		while (fgets(line, MAXLINE+1, pcfp) != NULL) {
			if (strstr(line, target) && line[0] != '#')
				continue;
			fputs(line, tpcfp);
		}
	}
	if ((an->state == ADD) || (an->aux && *(an->aux)))
		fprintf(tpcfp, "%s\n", an->aux);
	if (fclose(tpcfp)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in edit_piktd_conf(), fclose() error for %s", piktdconftmp);
		res = ERROR;
		return &res;
	}
	if (pcfp) {
		if (fclose(pcfp)) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in edit_piktd_conf(), fclose() error for %s", piktdconf);
			res = ERROR;
			return &res;
		}
		if ((res = remove(piktdconf)) == ERROR) {	/* remove returns 0 on success */
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in edit_piktd_conf(), remove() error for %s", piktdconf);
			return &res;
		}
	}
	if ((res = rename(piktdconftmp, piktdconf)) == ERROR) {	/* 0 success, -1 failure */
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in edit_piktd_conf(), rename() error for %s and %s", piktdconftmp, piktdconf);
		return &res;
	}
	chmod(piktdconf, (mode_t)(S_IRUSR|S_IWUSR|S_IRGRP));
	res = 0;
	return &res;
}

/* 
 */
int *
EXECUTE_COMMAND_1_SVC (ritm *cn, struct svc_req *rqstp)
{
	static char cmd[MAXLINE+1];
	pid_t pid;
	static int res;
	int status;

	if ((res = chkrequest(cn, (rfil *)NULL, rqstp, "execute_command")) != 0)
		return &res;

	if (cn->aux && *(cn->aux)) {
		if (strlen(cn->aux) > MAXLINE) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_command(), command string exceeds MAXLINE (%d) chars", MAXLINE);
			res = ERROR;
			return &res;
		} else
			strcpy(cmd, cn->aux);
	} else {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_command(), missing command string");
		res = ERROR;
		return &res;
	}

	switch (pid = fork()) {
		case -1:	/* error */
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE,  "in execute_command(), fork error");
			break;
		case 0:		/* child */
			logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "%s", cmd);
			execl(SHELL, "sh", "-c", cmd, (char *) NULL);
			_exit(127);
			break;
		default:	/* parent */
			waitpid(pid, &status, waitpid_options);
			break;
	}
	res = 0;
	return &res;
}

/* 
 */
execres *
EXECUTE_COMMANDWAIT_1_SVC (ritm *cn, struct svc_req *rqstp)
{
	static char cmd[MAXLINE+1];
	FILE *cmdproc;

	if ((xres.result = chkrequest(cn, (rfil *)NULL, rqstp, "execute_command")) != 0)
		return &xres;

	if (cn->aux && *(cn->aux)) {
		if (strlen(cn->aux)+5 > MAXLINE) {	/* +5 for "2>&1" */
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_commandwait(), command string exceeds MAXLINE (%d) chars", MAXLINE);
			xres.result = ERROR;
			return &xres;
		} else {
			/* check to see if "2>&1" already part of cn->aux? */
			sprintf(cmd, "%s 2>&1", cn->aux);
		}
	} else {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_commandwait(), missing command string");
		xres.result = ERROR;
		return &xres;
	}

	/* as a precaution, free execcmdout both before executing the
	   command, and after the master piktc is done acquiring the output
	   (when it sends a followup free_execcmdout() service request
	   to the client)
	 */
	free_execcmdout_svc();

	logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "%s", cmd);
	if ((cmdproc = popen(cmd, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_commandwait(), popen() error");
		xres.result = ERROR;
		return &xres;
	}
	xres.execres_u.execcmdout = (char *) mycalloc((size_t)(execcmdout_limit+1), sizeof(char), "execute_commandwait");
	if ((fread(xres.execres_u.execcmdout, sizeof(char), execcmdout_limit, cmdproc) == 0) && ferror(cmdproc)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_commandwait(), fread() error");
		xres.result = ERROR;
	} else
		xres.result = 0;
	pclose(cmdproc);
	return &xres;
}

/* just call free_execcmdout_svc(), always assuming a successful free
 */
int *
FREE_EXECCMDOUT_1_SVC (void *junk, struct svc_req *rqstp)
{
	static int res;

	free_execcmdout_svc();
	res = 0;
	return &res;
}

/* garbage collect the memory hunk previously allocated to temporary storage
   of exec_commandwait() output; this is a standalone function just in case
   this grows in complexity someday
 */
void
free_execcmdout_svc (void)
{
	FREE(xres.execres_u.execcmdout);
	xres.execres_u.execcmdout = (char *) NULL;
	return;
}

/* for the given alert, exec its corresponding confline (passed from the
   master, not read from the client piktd.conf)
 */
int *
EXECUTE_ALERT_1_SVC (ritm *an, struct svc_req *rqstp)
{
	static char *fn;
	struct stat fbuf;
	static char cmd[MAXLINE+1];
	register char *cp;
	register int i;
	int spcnt;
	pid_t pid;
	static int res;
	int status;

	if ((res = chkrequest(an, (rfil *)NULL, rqstp, "execute_alert")) != 0)
		return &res;

	fn = (char *) mymalloc((strlen(altdir)+strlen(an->name)+6) * sizeof(char), "EXECUTE_ALERT_1_SVC");
	sprintf(fn, "%s/%s.alt", altdir, an->name);
	if (stat(fn, &fbuf)) {	/* stat returns 0 on success,
				   -1 (ERROR) on failure
				 */
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_alert(), %s/%s.alt not found", altdir, an->name);
		res = ERROR;
		return &res;
	}

	if (an->aux && *(an->aux)) {
		cp = an->aux;
		spcnt = 0;
		/* count spaces until we get just past the 6 timing specs */
		while (*cp && (spcnt < 6)) {
			if (isspace((int)*cp))
				spcnt++;
			cp++;
		}
		if (! *cp || (spcnt < 6)) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_alert(), bad command string");
			res = ERROR;
			return &res;
		} else {
			i = 0;
			while (! isspace((int)*cp)) {
				cmd[i++] = *(cp++);
			}
			if (istest) {
				cmd[i++] = ' ';
				cmd[i++] = '-';
				cmd[i++] = 'T';
			}
			while (*cp) {
				cmd[i++] = *(cp++);
			}
			cmd[i] = '\0';
			strcat(cmd, " >/dev/null 2>&1");
		}
	} else {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in execute_alert(), missing command string");
		res = ERROR;
		return &res;
	}

	switch (pid = fork()) {
		case -1:	/* error */
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE,  "in execute_alert(), fork error");
			break;
		case 0:		/* child */
			logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "%s", cmd);
			execl(SHELL, "sh", "-c", cmd, (char *) NULL);
			_exit(127);
			break;
		default:	/* parent */
			waitpid(pid, &status, waitpid_options);
			break;
	}
	res = 0;
	return &res;
}

/* kill the piktd daemon
 */
int *
KILL_PIKTD_1_SVC (void *junk, struct svc_req *rqstp)
{
	pid_t oldpid, newpid;
	struct stat fbuf;
	char line[MAXLINE+1];
	FILE *fc;
	static int res;

	if ((res = chkrequest((ritm *)NULL, (rfil *)NULL, rqstp, "kill_piktd")) != 0)
		return &res;

	if ((oldpid = daemonpid()) == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in kill_piktd(), unable to kill piktd: multiple piktd's running");
		res = ERROR;
		return &res;
	} else if (oldpid == NONE) {
		res = NONE;
		return &res;
	}
	if (kill(oldpid, SIGKILL) != 0) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in kill_piktd(), failure of kill command");
		res = ERROR;
		return &res;
	}
	if ((newpid = daemonpid()) != NONE) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in kill_piktd(), unable to kill piktd (pid %d)", newpid);
		res = ERROR;
		return &res;
	} else
		logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "piktd (pid %d) killed", oldpid);
	/* killing piktd, and pikt runs, may yield a core file */
	if (lstat("/core", &fbuf) == 0) {
		if ((fc = popen(FILECMD " /core", "r")) == NULL) {
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in kill_piktd(), popen() error");
		} else {
			fgets(line, MAXLINE+1, fc);
			pclose(fc);
			logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in kill_piktd(), found /core file: %s", line);
		}
		if (remove("/core") == 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "rm'ed /core file");
	}
	/* remove the piktd lock file */
	if (lstat(piktdlock, &fbuf) == 0) {
		if (remove(piktdlock) != 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in kill_piktd(), unable to remove the %s file", piktdlock);
	}
	/* remove the piktd alerts file */
	if (lstat(piktdalt, &fbuf) == 0) {
		if (remove(piktdalt) != 0)	/* remove returns 0 on success */
			logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in kill_piktd(), unable to remove the %s file", piktdalt);
	}
	res = 0;
	return &res;
}

/* kill the piktc_svc daemon
 */
int *
KILL_PIKTC_SVC_1_SVC (void *junk, struct svc_req *rqstp)
{
	static int res;

	if ((res = chkrequest((ritm *)NULL, (rfil *)NULL, rqstp, "kill_piktc_svc")) != 0)
		return &res;
	/* unregister the service */
	svc_unregister(PIKTCPROG, PIKTCVERS);
	/* remove the lock */
	if (rmlocks(TRUE, TRUE) == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in kill_piktc_svc(), rmlocks() error");
		res = ERROR;
		return &res;
	}
	logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "daemon killed");
	/* no &res returned to caller */
	exit(0);
}

/* restart the piktd daemon
 */
int *
RESTART_PIKTD_1_SVC (void *junk, struct svc_req *rqstp)
{
	pid_t oldpid, newpid;
	pid_t pid;
	struct stat fbuf;
	static int res;
	int status;

	if ((res = chkrequest((ritm *)NULL, (rfil *)NULL, rqstp, "restart_piktd")) != 0)
		return &res;

	if ((oldpid = daemonpid()) == NONE) {
		/* if no piktd's running, remove any orphaned piktd lock
		   files
		 */
		if (lstat(piktdlock, &fbuf) == 0) {
			if (remove(piktdlock) != 0) {	/* remove returns 0 on success */
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in restart_piktd(), unable to restart piktd: unable to remove %s file", piktdlock);
				res = ERROR;
				return &res;
			}
		}
	} else if (oldpid == ERROR) { /* we have multiple piktd's running */
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in restart_piktd(), unable to restart piktd: multiple piktd's running");
		res = ERROR;
		return &res;
	} else { /* we have a running piktd */
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in restart_piktd(), unable to restart piktd: another piktd (pid %d) running", oldpid);
		res = ERROR;
		return &res;
	}
	switch (pid = fork()) {
		case -1:	/* error */
			logexit(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in restart_piktd(), fork() error");
			break;
		case 0:		/* child */
			execl(SHELL, "sh", "-c", piktd, (char *) NULL);
			_exit(127);
			break;
		default:	/* parent */
			waitpid(pid, &status, waitpid_options);
			break;
	}
	newpid = daemonpid();
	if ((newpid == ERROR) || (newpid == oldpid)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in restart_piktd(), failed to restart piktd");
		res = ERROR;
		return &res;
	}
	logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "piktd restart");
	res = 0;
	return &res;
}

/* restart the piktc_svc daemon; actually we only force the daemon to reread
   the PIKT.conf file
 */
int *
RESTART_PIKTC_SVC_1_SVC (void *junk, struct svc_req *rqstp)
{
	static char _RESTART_PIKTC_SVC_1_SVC_[] = "RESTART_PIKTC_SVC_1_SVC";
	static int res;

	if ((res = chkrequest((ritm *)NULL, (rfil *)NULL, rqstp, "restart_piktc_svc")) != 0)
		return &res;
	services = ALL_SERVICES;	/* re-enable all services before
					   rereading PIKT.conf
					 */
	getconf(homdir);
	FREE(hostserv);
	FREE(rqstserv);
	hostserv = mystrdup("", _RESTART_PIKTC_SVC_1_SVC_);
	rqstserv = mystrdup("", _RESTART_PIKTC_SVC_1_SVC_);
	statserv = NONE;
	logreturn(PIKTLOG, LOG_INFO, __FILE__, __LINE__, FALSE, "daemon restart");
	res = 0;
	return &res;
}

/* report the status of the given alert; an alert is inactive if: there is
   no .alt file, there is no line for that alert in piktd.conf, there is
   no active piktd, and/or the alert is absent from the piktd alerts file
 */
int *
SHOW_ALERT_STATUS_1_SVC (ritm *an, struct svc_req *rqstp)
{
	FILE *pcfp, *rfp;
	static char *fn;
	char line[MAXLINE+1];
	char target[MAXLINE+1];
	struct stat fbuf;
	static int res;
	char *cp;

	if ((res = chkrequest(an, (rfil *)NULL, rqstp, "show_alert_status")) != 0)
		return &res;

	/* is there an .alt file? */
	fn = (char *) mymalloc((strlen(altdir)+strlen(an->name)+6) * sizeof(char), "SHOW_ALERT_STATUS_1_SVC");
	sprintf(fn, "%s/%s.alt", altdir, an->name);
	if ((res = stat(fn, &fbuf)) == ERROR) {	/* 0 success, -1 failure */
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in show_alert_status(), stat() error for %s", fn);
		free(fn);
		return &res;
	}

	/* found in piktd.conf? */
	if ((pcfp = fopen(piktdconf, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in show_alert_status(), fopen() error for %s", piktdconf);
		res = ERROR;
		return &res;
	}
	sprintf(target, "+A %s", an->name);
	res = ERROR;
	while (fgets(line, MAXLINE+1, pcfp) != NULL) {
		if (strstr(line, target) && line[0] != '#') {
			res = 0;
			break;
		}
	}
	if (fclose(pcfp)) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in show_alert_status(), fclose() error for %s", piktdconf);
		res = ERROR;
		return &res;
	}
	if (res == ERROR) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in show_alert_status(), %s entry not found in %s", an->name, piktdconf);
		return &res;
	}

	/* active piktd? */
	if ((res = daemonpid()) == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in show_alert_status(), active piktd not found");
		return &res;
	}

	/* registered in active pikd? */
	if ((rfp = fopen(piktdalt, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, TRUE, "in show_alert_status(), fopen() error for %s", piktdalt);
		res = ERROR;
		return &res;
	}
	res = ERROR;
	while (fgets(line, MAXLINE+1, rfp) != NULL) {
		if ((cp = strchr(line, '\n')) != NULL)
			*cp = '\0';
		if (! strcmp(line, an->name)) {
			res = 0;
			break;
		}
	}
	if (res == ERROR) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in show_alert_status(), %s not found in %s", an->name, piktdalt);
		return &res;
	}

	return &res;
}

/* return the piktd pid, or NONE if no daemon is running, or ERROR if
   multiple piktd's are running; when determining this, ignore piktd's
   spawned for launching pikt runs

   we used to invoke ioctl() on each /proc/<pid> to get the process names,
   but we saw several cases where the resultant process name for a given
   pid didn't match the process name for that pid as shown in the ps display;
   the technique of reading the ps display, although perhaps less elegant,
   gives us the consistent results we desire
 */
pid_t
daemonpid (void)
{
	pid_t pspid;
	long pid, ppid;
	#if PSPARSE == 1
	char psp[MAXLINE+1];
	#endif
	char *psproc;
	char line[MAXLINE+1];
	FILE *ps;
	char *cp;

	pspid = 0;
	if ((ps = popen(PSCMD " " PSOPTS, "r")) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "in daemonpid(), popen() error");
		return ERROR;
	}
	fgets(line, MAXLINE+1, ps);	/* throw away header */
	while (fgets(line, MAXLINE+1, ps)) {
		#if PSPARSE == 1
		if (sscanf(line, "%ld %ld %s", &pid, &ppid, psp) != 3) {
			logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in daemonpid(), error reading pid/ppid/psproc for line %s", line);
			continue;
		}
		/* strip away path, if any, leaving only basename */
		if ((cp = strrchr(psp, '/')))
			psproc = cp+1;
		else
			psproc = psp;
		#else
		# if PSPARSE == 2
		if (sscanf(line, "%*s %*s %ld %ld %*s %*s %*s %*s %*s %*s %*s %*s %*s", &pid, &ppid) != 2) {
			logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in daemonpid(), error reading pid/ppid for line %s", line);
			continue;
		}
		if ((psproc = strrchr(line, ':')))
			psproc += 4;
		else {
			logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in daemonpid(), error reading psproc for line %s", line);
			continue;
		}
		# else
		#  if PSPARSE == 3
		if (sscanf(line, "%*s %*s %*s %ld %ld %*s %*s %*s %*s %*s %*s %*s %*s %*s", &pid, &ppid) != 2) {
			logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in daemonpid(), error reading pid/ppid for line %s", line);
			continue;
		}
		if ((psproc = strrchr(line, ':')))
			psproc += 4;
		else {
			logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "in daemonpid(), error reading psproc for line %s", line);
			continue;
		}
		#  else
		#   include "Bletch: How do we parse `ps' output?"
		#  endif
		# endif
		#endif
		/* for some strange reason, linux ps appends a space to the
		   end of each ps line
		 */
		if ((! strcmp(psproc, "piktd") || ! strncmp(psproc, "piktd\n", 6) || ! strncmp(psproc, "piktd ", 6)) && (ppid == 1)) {
			if (pspid == 0)
				pspid = (pid_t) pid;
			else {
				logreturn(PIKTLOG, LOG_ERR, __FILE__, __LINE__, FALSE, "failed to determine unique piktd pid: multiple piktd's running");
				pclose(ps);
				return ERROR;
			}
		}
	}
	pclose(ps);
	if (pspid == 0) {
		logreturn(PIKTLOG, LOG_WARNING, __FILE__, __LINE__, FALSE, "failed to determine piktd pid: piktd not running");
		return NONE;
	} else
		return pspid;
}
